@@ -2,11 +2,15 @@ from __future__ import absolute_import
 from __future__ import division
 from __future__ import print_function
 
+__all__ = ["Model", "TrainState", "LossHistory"]
+
+
 import pickle
 from collections import OrderedDict
 
 import numpy as np
 
+from . import config
 from . import display
 from . import gradients as grad
 from . import losses as losses_module
@@ -18,11 +22,11 @@ from .callbacks import CallbackList
 
 
 class Model(object):
-    """A ``Model`` trains a ``Map`` on a ``Data``.
+    """A ``Model`` trains a ``NN`` on a ``Data``.
 
     Args:
         data: ``deepxde.data.Data`` instance.
-        net: ``deepxde.maps.Map`` instance.
+        net: ``deepxde.nn.NN`` instance.
     """
 
     def __init__(self, data, net):
@@ -40,7 +44,7 @@ class Model(object):
 
         # Backend-dependent attributes
         # Tensor or callable
-        self.losses = None
+        self.outputs = None
         self.outputs_losses = None
         self.train_step = None
         if backend_name == "tensorflow.compat.v1":
@@ -136,7 +140,7 @@ class Model(object):
         total_loss = tf.math.reduce_sum(losses)
 
         # Tensors
-        self.losses = losses
+        self.outputs = self.net.outputs
         self.outputs_losses = [self.net.outputs, losses]
         self.train_step = optimizers.get(
             total_loss, self.opt_name, learning_rate=lr, decay=decay
@@ -145,9 +149,24 @@ class Model(object):
     def _compile_tensorflow(self, lr, loss_fn, decay, loss_weights):
         """tensorflow"""
 
-        def compute_losses(targets, outputs):
+        # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
+        @tf.function
+        def outputs(training, inputs):
+            return self.net(inputs, training=training)
+
+        # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
+        @tf.function
+        def outputs_losses(training, inputs, targets, auxiliary_vars):
+            # TODO: Add training
+            # self.net.training = training
+            self.net.inputs = inputs
+            self.net.targets = targets
+            self.net.auxiliary_vars = auxiliary_vars
+            # Don't call outputs() decorated by @tf.function above, otherwise the
+            # gradient of outputs wrt inputs will be lost here.
+            outputs_ = self.net(inputs, training=training)
             # Data losses
-            losses = self.data.losses(targets, outputs, loss_fn, self)
+            losses = self.data.losses(targets, outputs_, loss_fn, self)
             if not isinstance(losses, list):
                 losses = [losses]
             # Regularization loss
@@ -155,32 +174,19 @@ class Model(object):
                 losses += [tf.math.reduce_sum(self.net.losses)]
             losses = tf.convert_to_tensor(losses)
             # TODO: Weighted losses
-            return losses
-
-        # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
-        @tf.function
-        def outputs_losses(training, data_id, inputs, targets, auxiliary_vars=None):
-            self.net.data_id = data_id
-            self.net.inputs = inputs
-            self.net.targets = targets
-            self.net.auxiliary_vars = auxiliary_vars
-            outputs = self.net(inputs, training=training)
-            losses = compute_losses(targets, outputs)
-            return outputs, losses
+            if loss_weights is not None:
+                raise NotImplementedError(
+                    "Backend tensorflow doesn't support loss_weights"
+                )
+            return outputs_, losses
 
         opt = optimizers.get(self.opt_name, learning_rate=lr, decay=decay)
 
         @tf.function
-        def train_step(training, data_id, inputs, targets, auxiliary_vars=None):
-            # inputs and targets are np.ndarray, and automatically converted to
-            # tf.Tensor without memory copy:
-            # https://www.tensorflow.org/tutorials/customization/basics#numpy_compatibility
-            # But, this doesn't seem to be true:
-            # https://github.com/tensorflow/tensorflow/issues/33254
+        def train_step(inputs, targets, auxiliary_vars):
+            # inputs and targets are np.ndarray and automatically converted to Tensor.
             with tf.GradientTape() as tape:
-                _, losses = outputs_losses(
-                    training, data_id, inputs, targets, auxiliary_vars
-                )
+                losses = outputs_losses(True, inputs, targets, auxiliary_vars)[1]
                 total_loss = tf.math.reduce_sum(losses)
             trainable_variables = (
                 self.net.trainable_variables + self.external_trainable_variables
@@ -188,51 +194,115 @@ class Model(object):
             grads = tape.gradient(total_loss, trainable_variables)
             opt.apply_gradients(zip(grads, trainable_variables))
 
+        def train_step_tfp(inputs, targets, auxiliary_vars):
+            def build_loss():
+                losses = outputs_losses(True, inputs, targets, auxiliary_vars)[1]
+                return tf.math.reduce_sum(losses)
+
+            trainable_variables = (
+                self.net.trainable_variables + self.external_trainable_variables
+            )
+            return opt(trainable_variables, build_loss)
+
         # Callables
-        self.losses = compute_losses
+        self.outputs = outputs
         self.outputs_losses = outputs_losses
-        self.train_step = train_step
+        self.train_step = (
+            train_step
+            if not optimizers.is_external_optimizer(self.opt_name)
+            else train_step_tfp
+        )
 
     def _compile_pytorch(self, lr, loss_fn, decay, loss_weights):
         """pytorch"""
 
-        def compute_losses(targets, outputs):
+        def outputs(inputs):
+            # TODO: Add training
+            # TODO: Use torch.no_grad() if training is False
+            inputs = torch.as_tensor(inputs)
+            inputs.requires_grad_()
+            self.net.inputs = inputs
+            return self.net(inputs)
+
+        def outputs_losses(inputs, targets):
+            outputs_ = outputs(inputs)
             # Data losses
-            losses = self.data.losses(targets, outputs, loss_fn, self)
+            if targets is not None:
+                targets = torch.as_tensor(targets)
+            losses = self.data.losses(targets, outputs_, loss_fn, self)
             if not isinstance(losses, list):
                 losses = [losses]
             # TODO: regularization
             # TODO: Weighted losses
+            if loss_weights is not None:
+                raise NotImplementedError(
+                    "Backend pytorch doesn't support loss_weights"
+                )
             losses = torch.stack(losses)
             # Clear cached Jacobians and Hessians.
             grad.clear()
-            return losses
-
-        def outputs_losses(inputs, targets):
-            inputs = torch.from_numpy(inputs)
-            targets = torch.from_numpy(targets)
-            inputs.requires_grad_()
-            self.net.inputs = inputs
-            outputs = self.net(inputs)
-            losses = compute_losses(targets, outputs)
-            return outputs, losses
+            return outputs_, losses
 
+        # Another way is using per-parameter options
+        # https://pytorch.org/docs/stable/optim.html#per-parameter-options,
+        # but not all optimizers (such as L-BFGS) support this.
+        trainable_variables = (
+            list(self.net.parameters()) + self.external_trainable_variables
+        )
         opt = optimizers.get(
-            self.net.parameters(), self.opt_name, learning_rate=lr, decay=decay
+            trainable_variables, self.opt_name, learning_rate=lr, decay=decay
         )
 
         def train_step(inputs, targets):
-            _, losses = outputs_losses(inputs, targets)
-            total_loss = torch.sum(losses)
-            opt.zero_grad()
-            total_loss.backward()
-            opt.step()
+            def closure():
+                losses = outputs_losses(inputs, targets)[1]
+                total_loss = torch.sum(losses)
+                opt.zero_grad()
+                total_loss.backward()
+                return total_loss
+
+            opt.step(closure)
 
         # Callables
-        self.losses = compute_losses
+        self.outputs = outputs
         self.outputs_losses = outputs_losses
         self.train_step = train_step
 
+    def _outputs(self, training, inputs):
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(training, inputs)
+            return self.sess.run(self.outputs, feed_dict=feed_dict)
+        if backend_name == "tensorflow":
+            outs = self.outputs(training, inputs)
+        elif backend_name == "pytorch":
+            # TODO: training
+            with torch.no_grad():
+                outs = self.outputs(inputs)
+        return utils.to_numpy(outs)
+
+    def _train_step(self, inputs, targets, auxiliary_vars):
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(True, inputs, targets, auxiliary_vars)
+            self.sess.run(self.train_step, feed_dict=feed_dict)
+        elif backend_name == "tensorflow":
+            self.train_step(inputs, targets, auxiliary_vars)
+        elif backend_name == "pytorch":
+            # TODO: auxiliary_vars
+            self.train_step(inputs, targets)
+
+    def _run(self, fetches, training, inputs, targets, auxiliary_vars):
+        """Runs one "step" of computation of tensors or callables in `fetches`."""
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(training, inputs, targets, auxiliary_vars)
+            return self.sess.run(fetches, feed_dict=feed_dict)
+        if backend_name == "tensorflow":
+            outs = fetches(training, inputs, targets, auxiliary_vars)
+        elif backend_name == "pytorch":
+            # TODO: Use torch.no_grad() in _test() and predict()
+            # TODO: training, auxiliary_vars
+            outs = fetches(inputs, targets)
+        return None if outs is None else utils.to_numpy(outs)
+
     @utils.timing
     def train(
         self,
@@ -247,7 +317,8 @@ class Model(object):
         """Trains the model for a fixed number of epochs (iterations on a dataset).
 
         Args:
-            epochs: Integer. Number of epochs to train the model.
+            epochs: Integer. Number of iterations to train the model. Note: It is the
+                number of iterations, not the number of epochs.
             batch_size: Integer or ``None``. If you solve PDEs via ``dde.data.PDE`` or
                 ``dde.data.TimePDE``, do not use `batch_size`, and instead use
                 `dde.callbacks.PDEResidualResampler
@@ -286,7 +357,10 @@ class Model(object):
         self._test()
         self.callbacks.on_train_begin()
         if optimizers.is_external_optimizer(self.opt_name):
-            self._train_scipy(display_every)
+            if backend_name == "tensorflow.compat.v1":
+                self._train_tensorflow_compat_v1_scipy(display_every)
+            elif backend_name == "tensorflow":
+                self._train_tensorflow_tfp()
         else:
             if epochs is None:
                 raise ValueError("No epochs for {}.".format(self.opt_name))
@@ -307,10 +381,7 @@ class Model(object):
             self.train_state.set_data_train(
                 *self.data.train_next_batch(self.batch_size)
             )
-            self._run(
-                self.train_step,
-                True,
-                np.uint8(0),
+            self._train_step(
                 self.train_state.X_train,
                 self.train_state.y_train,
                 self.train_state.train_aux_vars,
@@ -327,8 +398,7 @@ class Model(object):
             if self.stop_training:
                 break
 
-    def _train_scipy(self, display_every):
-        # TODO: backend tensorflow
+    def _train_tensorflow_compat_v1_scipy(self, display_every):
         def loss_callback(loss_train):
             self.train_state.epoch += 1
             self.train_state.step += 1
@@ -344,7 +414,6 @@ class Model(object):
         self.train_state.set_data_train(*self.data.train_next_batch(self.batch_size))
         feed_dict = self.net.feed_dict(
             True,
-            0,
             self.train_state.X_train,
             self.train_state.y_train,
             self.train_state.train_aux_vars,
@@ -352,16 +421,27 @@ class Model(object):
         self.train_step.minimize(
             self.sess,
             feed_dict=feed_dict,
-            fetches=[self.losses],
+            fetches=[self.outputs_losses[1]],
             loss_callback=loss_callback,
         )
         self._test()
 
+    def _train_tensorflow_tfp(self):
+        self.train_state.set_data_train(*self.data.train_next_batch(self.batch_size))
+        results = self.train_step(
+            self.train_state.X_train,
+            self.train_state.y_train,
+            self.train_state.train_aux_vars,
+        )
+        # TODO: Loss history
+        self.train_state.epoch += results.num_iterations.numpy()
+        self.train_state.step += results.num_iterations.numpy()
+        self._test()
+
     def _test(self):
         self.train_state.y_pred_train, self.train_state.loss_train = self._run(
             self.outputs_losses,
-            False,
-            np.uint8(0),
+            True,
             self.train_state.X_train,
             self.train_state.y_train,
             self.train_state.train_aux_vars,
@@ -369,7 +449,6 @@ class Model(object):
         self.train_state.y_pred_test, self.train_state.loss_test = self._run(
             self.outputs_losses,
             False,
-            np.uint8(1),
             self.train_state.X_test,
             self.train_state.y_test,
             self.train_state.test_aux_vars,
@@ -398,25 +477,21 @@ class Model(object):
 
     def predict(self, x, operator=None, callbacks=None):
         """Generates output predictions for the input samples."""
+        x = x.astype(config.real(np))
         self.callbacks = CallbackList(callbacks=callbacks)
         self.callbacks.set_model(self)
         self.callbacks.on_predict_begin()
-        # TODO: predict operator with auxiliary_vars
-        if backend_name == "tensorflow.compat.v1":
-            if operator is None:
-                op = self.net.outputs
-            else:
+        if operator is None:
+            y = self._outputs(False, x)
+        else:
+            # TODO: predict operator with auxiliary_vars
+            if backend_name == "tensorflow.compat.v1":
                 if utils.get_num_args(operator) == 2:
                     op = operator(self.net.inputs, self.net.outputs)
                 elif utils.get_num_args(operator) == 3:
                     op = operator(self.net.inputs, self.net.outputs, x)
-            y = self._run(op, False, np.uint8(2), x, None, None)
-        elif backend_name == "tensorflow":
-            # TODO: avoid creating the same graph every time predict is called
-            # TODO: use self._run for tensorflow
-            if operator is None:
-                y = self.net(x).numpy()
-            else:
+                y = self._run(op, False, x, None, None)
+            elif backend_name == "tensorflow":
                 if utils.get_num_args(operator) == 2:
 
                     @tf.function
@@ -431,13 +506,17 @@ class Model(object):
                         y = self.net(inputs)
                         return operator(inputs, y, x)
 
-                y = op(x).numpy()
-        elif backend_name == "pytorch":
-            # TODO
-            raise NotImplementedError(
-                "Model.predict hasn't been implemented for backend pytorch."
-            )
-
+                y = op(x)
+                y = utils.to_numpy(y)
+            elif backend_name == "pytorch":
+                inputs = torch.as_tensor(x)
+                inputs.requires_grad_()
+                outputs = self.net(inputs)
+                if utils.get_num_args(operator) == 2:
+                    y = operator(inputs, outputs)
+                elif utils.get_num_args(operator) == 3:
+                    y = operator(inputs, outputs, x)
+                y = utils.to_numpy(y)
         self.callbacks.on_predict_end()
         return y
 
@@ -450,6 +529,10 @@ class Model(object):
     def state_dict(self):
         """Returns a dictionary containing all variables."""
         # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         destination = OrderedDict()
         variables_names = [v.name for v in tf.global_variables()]
         values = self.sess.run(variables_names)
@@ -467,6 +550,10 @@ class Model(object):
                 "tf.train.Saver" protocol supports ``restore()``.
         """
         # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         if verbose > 0:
             print(
                 "Epoch {}: saving model to {}-{} ...\n".format(
@@ -482,6 +569,10 @@ class Model(object):
     def restore(self, save_path, verbose=0):
         """Restore all variables from a disk file."""
         # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         if verbose > 0:
             print("Restoring model from {} ...\n".format(save_path))
         self.saver.restore(self.sess, save_path)
@@ -489,31 +580,16 @@ class Model(object):
     def print_model(self):
         """Prints all trainable variables."""
         # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         variables_names = [v.name for v in tf.trainable_variables()]
         values = self.sess.run(variables_names)
         for k, v in zip(variables_names, values):
             print("Variable: {}, Shape: {}".format(k, v.shape))
             print(v)
 
-    def _run(self, fetches, training, data_id, inputs, targets, auxiliary_vars):
-        """Runs one "step" of computation of tensors or callables in `fetches`."""
-        if backend_name == "tensorflow.compat.v1":
-            feed_dict = self.net.feed_dict(
-                training,
-                data_id,
-                inputs,
-                targets,
-                auxiliary_vars,
-            )
-            return self.sess.run(fetches, feed_dict=feed_dict)
-        if backend_name == "tensorflow":
-            outs = fetches(training, data_id, inputs, targets, auxiliary_vars)
-            return None if outs is None else [out.numpy() for out in outs]
-        if backend_name == "pytorch":
-            # TODO: Use torch.no_grad() in _test() and predict()
-            outs = fetches(inputs, targets)
-            return None if outs is None else [out.detach().numpy() for out in outs]
-
 
 class TrainState(object):
     def __init__(self):
